@@ -22,10 +22,12 @@ elsif (-e "./buildenv.pl")
 }
 
 my $target = shift || Usage();
-Install($target);
+my $insttype = shift;
+Install($target, $insttype);
 
 sub Usage
 {
-	print "Usage: install.pl <targetdir>\n";
+	print "Usage: install.pl <targetdir> [installtype]\n";
+	print "installtype: client\n";
 	exit(1);
 }
